package provide mmucl 0.99.9

##########################################################################
# mmucl.tcl - The core mmucl library.
#
# Copyright (C) 1997 Mark Patton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##########################################################################

# Contents of global array Mmucl:
#  connection - whether or not connected to mud
#  histloc    - location of input given to mmucl_parse in history
#  host       - mud host
#  port       - mud port
#  char       - current character loaded by LoadChar if any
#  sockid     - socket identifier for connection to mud
#  rc_dir     - location of users charlist and character init files.
#  lib_dir    - location of libraries, help files etc.
#  browser    - html browser to use for viewing html help files.
#  reconnect  - if 1 reconnect after timeout. if 2 reconnect after
# any disconnect. if 0 don't reconnect

# Some elements need to be set.
# rc_dir and lib_dir should be set by program using the library.
# The other elements will be set by various procedures.

array set Mmucl {
    connection 0
    histloc 0
}

# Internal procs that should not be called from the command line.
#################################################################

# echo and display are callbacks that need to be defined.
#  display displays mud output and may do special formatting.
#  echo just displays the given string.

# get output from mud, check it for actions and display it.
# if an action does match do substitution for @n
proc mmucl_read {} {
    global MmuclAction
    upvar \#0 Mmucl(sockid) sockid

    if {[catch {set mudout [read $sockid]} error] || [eof $sockid]} {
	if {$error!=""} {echo "** $error **"}
	disconnect; return
    }
    display $mudout

    foreach {pattern cmds} [array get MmuclAction] {
	if {[regexp -nocase $pattern $mudout 0 1 2 3 4 5 6 7 8 9]} {
	    regsub -all {\$} [join $cmds] {\\&} cmds
	    regsub -all {([^\\])%([0-9])} $cmds {\1$\2} cmds
	    if {[catch {uplevel \#0 [subst -nocommands $cmds]} error]!=0} {
		echo "** Action \{$pattern\} error: $error **"
	    }
	}
    }
}

# Check input to see what should be done with it.
#  \ means send it to the mud
#  # means input is a tcl command.
# input is considered to consist of commands seperated
# by ;. Each command is checked to see if it matches an
# alias and if so substitution of n arg for %n is done.

# Have to do a special case for "" (return) because
# the split eats it up. Also because return is "" the history
# command ignores it...

proc mmucl_parse {str} {
    upvar \#0 Mmucl(histloc) histloc
    
    set fchar [string index $str 0]
    if {$fchar=="\\"} {
	mudputs [string trimleft $str \\]
    } elseif {$fchar=="#"} {
	if {[catch {uplevel \#0 [string trimleft $str \#]} error]} {
	    echo "** Error: $error **"
	}
    } elseif {$str==""} {
 	mudputs ""
    } else {
	foreach cmd [split $str \;] {
	    regexp {([^ ]*) ?(.*)} $cmd x alias args
	    if {![exec_alias $alias $args]} {
		mudputs $cmd
	    }
	}
    }
    history add $str
    set histloc [history nextid]
}

# Putting the below in the foreach loop enables escaping
# a ; with \;. But I'm not sure its needed.
# if {[string match {*\\} $cmd]} {
#      append line [string trimright $cmd \\]\;
#      continue
#  }
# if {[info exists line]} {
#     append line $cmd; set cmd $line; unset line
# }


# These are procs, some of which are used internally,
# which can be called by the user.
#######################################################

# connect to a mud. If given the string login is written
# to the mud on a successful connect.
# Connect called with no args will try to resume a previous
# connection
proc connect {host port {login ""}} {
    global Mmucl MmuclChar

    if {$Mmucl(connection)} {
	echo "** Already connected. **"
	return
    }

    echo "** Connecting to $host at port $port... **"
    
    # Do the close and after cancel to make sure
    # we're only attempting one connection at a time.
    catch {close $Mmucl(sockid); after cancel $Mmucl(timeout_id)}

    # Set up a timeout.
    set Mmucl(timeout_id) [after 20000 {
	if {!$Mmucl(connection)} {
	    catch {close $Mmucl(sockid)}
	    echo "** Timed out. **"
	    if {[info exists Mmucl(reconnect)] && $Mmucl(reconnect)} {
		after 2000 reconnect
	    }
	}}]
    
    if {[catch {set Mmucl(sockid) [socket -async $host $port]} error]} {
	after cancel $Mmucl(timeout_id)
	echo "** Error: $error **"
    } else {
	fconfigure $Mmucl(sockid) -buffering line -blocking 0 \
	    -translation binary
	
	set Mmucl(host) $host
	set Mmucl(port) $port
	
	# Fileevent only gets called once a connection.
	# It sets a variable the after commands check to see if we
	# connect in a specific period of time and time out if we don't..
	
	fileevent $Mmucl(sockid) readable "
              after cancel $Mmucl(timeout_id); unset Mmucl(timeout_id)
              set Mmucl(connection) 1
              fileevent $Mmucl(sockid) readable mmucl_read
	      if {[string length $login]} {catch {writemud $login}}"
    }
}

# disconnect from a mud
proc disconnect {} {
    global Mmucl

    if {!$Mmucl(connection)} {
	echo "** Not connected **"
    } else {
	catch {close $Mmucl(sockid)}
	set Mmucl(connection) 0
	echo "** Connection closed. **"
	if {[info exists Mmucl(reconnect)] && $Mmucl(reconnect)==2} {
	    after 2000 reconnect
	}
    }
}

# try to resume previous connection. Mmucl(char) takes precedence.
proc reconnect {} {
    global Mmucl MmuclChar

    if {[info exists Mmucl(host)] && [info exists Mmucl(port)]} {
	set login ""
	catch {set login [lindex $MmuclChar($Mmucl(char)) 3]}
	connect $Mmucl(host) $Mmucl(port) $login
    } else {
	echo "** No previous connection **"
    }
}


# break up str into commands seperated by ; and
# write them to the mud.
proc writemud {str} {
    upvar \#0 Mmucl(sockid) sockid Mmucl(connection) connect
    
    if {!$connect} {
	echo "** Not connected. **"
    } else {
	foreach cmd [split $str \;] {
	    puts $sockid $cmd
	}
	flush $sockid
    }
}

# write str to mud
proc mudputs {str} {
    upvar \#0 Mmucl(sockid) sockid Mmucl(connection) connect
    
    if {!$connect} {
	echo "** Not connected. **"
    } else {
	puts $sockid $str
	flush $sockid
    }
}

# add or display aliases
proc alias {{name *} {str ""}} {
    global MmuclAlias
    
    if {$str==""} {
	set result [array names MmuclAlias $name]
	if {$result !=""} {
	    foreach match [lsort $result] {
		echo "** Alias $match: $MmuclAlias($match) **"
	    }
	} else {
	    echo "** No alias matching: $name **"
	}
    } else {
	set MmuclAlias($name) [list $str]
	echo "** Alias \"$name\" set to \{$str\} **"
    }
}

# delete an alias
proc unalias {name} {
    global MmuclAlias
    
    set result [array names MmuclAlias $name]
    if {$result !=""} {
	foreach match $result {
	    unset MmuclAlias($match)
	    echo "** \"$match\" is no longer an alias. **"
	}
    } else {
	echo "** No alias matching: $name. **"
    }
}

# add or display actions
proc action {{pattern *} {cmds ""}} {
    global MmuclAction
    
    if {$cmds==""} {
	set result [array names MmuclAction $pattern]
	if {$result !=""} {
	    foreach match [lsort $result] {
		echo "** Action \{$match\}: $MmuclAction($match) **"
	    }
	} else {
	    echo "** No action matching: $pattern **"
	}
    } else {
	# Check to see if the pattern is a valid regexp.
	if {[catch {regexp $pattern ""} error] !=0} {
	    echo "** Warning: $error"
	}
	set MmuclAction($pattern) [list $cmds]
	echo "** Action \{$pattern\} triggers \{$cmds\} **"
    }
}

# delete an action
proc unaction {pattern} {
    global MmuclAction

    # make sure pattern doesn't blow up in [array names...]
    regsub -all {[\[]} $pattern {\\&} pattern
		  
     set result [array names MmuclAction $pattern]
     if {$result !=""} {
       foreach match $result {
       unset MmuclAction($match)
	 echo "** \{$match\} is no longer an action. **"
	}
    } else {
	echo "** No action matching: $pattern. **"
    }
}
	
# exec an alias and do the substitutions of args for %n
proc exec_alias {name {arg {}}} {
    global MmuclAlias

    if {![info exists MmuclAlias($name)]} {
	return 0
    }
    regsub -all %0 [join $MmuclAlias($name)] [join $arg] cmds
    regsub -all {\[} $cmds {\\&} cmds
    regsub -all {([^\\])%([1-9])} $cmds \
	"\\1\[lindex [list $arg] \[expr {\\2 -1}]]" cmds  
    if {[catch {uplevel \#0 [subst -novariables $cmds]} error]!=0} {
	echo "** Alias $name error: $error **"
    }
    return 1
}

# Add a character to the file charlist in rc_dir
proc charadd {character mudname host port {login ""}} {
    global MmuclChar
    upvar \#0 Mmucl(rc_dir) rc_dir

    if {[file exists [file join $rc_dir charlist]]} {
	source [file join $rc_dir charlist]
    }

    set MmuclChar($character) [list $mudname $host $port $login]
    set fileid [open [file join $rc_dir charlist] w]
    puts $fileid [list array set MmuclChar [array get MmuclChar]] 
    flush $fileid
    close $fileid
    
    # Don't overwrite an init file if it's already there.
    if {[catch {set fileid [open [file join $rc_dir $character.$mudname]\
				{WRONLY CREAT}]} error]!=0} {
	echo "** Error: $error **"
	return
    }
    close $fileid

    echo "** Created character $character. **"
    echo "** Init file: [file join $rc_dir $character.$mudname] **"   
}

# Connect to defined character's mud and load his init file.
proc charload {name} {
     upvar \#0 Mmucl(rc_dir) rc_dir Mmucl(char) char
    
    if {![file exists [file join $rc_dir charlist]]} {
	echo "** No characters defined. **"
	return
    }
    source [file join $rc_dir charlist]
    if {![info exists MmuclChar($name)]} {
	echo "** No character \"$name\" defined. **"
	return
    }
    set charlist $MmuclChar($name)
    set init_file [file join $rc_dir $name.[lindex $charlist 0]]
    
    if {[file exists $init_file]} {
	source $init_file 
    } else {
	echo "** Init file $init_file not found **"
    }
    connect [lindex $charlist 1] [lindex $charlist 2] [lindex $charlist 3]
    set char $name
}

# delete a character
proc chardelete {name} {
    global MmuclChar
    upvar \#0 Mmucl(rc_dir) rc_dir

    if {![file exists [file join $rc_dir charlist]]} {
	echo "** No characters defined. **"
	return
    }
    source [file join $rc_dir charlist]

    if {![info exists MmuclChar($name)]} {
	echo "** No defined character $name. **"
	return
    }

    source [file join $rc_dir charlist]
    set charlist $MmuclChar($name)
    unset MmuclChar($name)
    set fileid [open [file join $rc_dir charlist] w]
    puts $fileid [list array set MmuclChar [array get MmuclChar]] 
    flush $fileid; close $fileid
    catch {file delete [file join $rc_dir $name.[lindex $charlist 0]]}
    echo "** Character $name deleted. **"
}

# pop up a local html browser.
proc help {{subject contents}} {
    upvar \#0 Mmucl(lib_dir) lib_dir Mmucl(browser) browser
    
    set file [file join $lib_dir html $subject.html]
    if {![file exists $file]} {
 	# should go to some kind of error page
 	set file [file join $lib_dir html contents.html]
    }
    regsub %1 $browser $file script
    
    # Fix up windows paths.
    regsub -all {\\} $script / script

    eval exec $script &
}

# save currently defined aliases and actions to a file.
proc save {filename} {
    global MmuclAlias MmuclAction

    if {[catch {set fileid [open $filename w]} error]!=0} {
	echo "** Error: $error **"
	return
    }
    foreach alias [array names MmuclAlias] {
	puts $fileid "alias \{$alias\} $MmuclAlias($alias)"
    }
    foreach action [array names MmuclAction] {
	puts $fileid "action \{$action\} $MmuclAction($action)"
    }
    flush $fileid
    close $fileid
    echo "** All defined aliases and actions saved in $filename. **"
}

# read a file directly into the mud.
proc textin {filename} {
    if {[catch {set fileid [open $filename r]} error]!=0} {
	echo "** $error **"
    } else {
	mudputs [read $fileid]
	close $fileid

    }
}