#########################################################################
# mmuclGui.tcl - graphical interfaces to core mmucl procedures
#
# Copyright (C) 1997 Mark Patton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#########################################################################

package provide mmuclGui 0.3

# procedures for making dialog boxes to edit arrays
###################################################

# put the elements of list in form w
proc form_put {w list} {
    set count 0
    foreach element $list {
        $w.f.e$count delete 0 end  
        $w.f.e$count insert 0 $element
        incr count
    }
}

# returns a list of the current values in w's entry widgets
proc form_get {w} {
    set number [expr {[llength [winfo children $w.f]]/2}]
    
    for {set count 0} {$count < $number} {incr count} {
        lappend result [$w.f.e$count get]
    }
    return $result
}

# delete a record of form w with the callback command.
# the delete command is given a index in the record's array
# to delete.
proc formMenu_delete {w command array_var} {
    set i [$w.f.sel.list curselection]
    if {$i!=""} {
	$command [$w.f.sel.list get $i]
	formMenu_update $w $array_var
    }
}

# save the record of form name with a callback command.
proc form_save {w formMenu_w array_var command} {
    set list [form_get $w]
    if {[lindex $list 0]!={}} {
        eval $command $list
        formMenu_update $formMenu_w $array_var
	form_active $w $formMenu_w $w
    }
}

# fill formMenu with array
proc formMenu_update {w array_var} {
    global $array_var
    set w $w.f.sel.list

    $w delete 0 end
    foreach element [lsort [array names $array_var]] {
        $w insert end $element
    }
}

# fills the form with the value from formMenu w
proc form_active {w formMenu_w array_var} {
    upvar \#0 $array_var array
    set lb $formMenu_w.f.sel.list
    
    set i [$lb curselection]
    if {$i!=""} {
	set list "\"[$lb get $i]\" $array([$lb get $i])"
	form_put $w $list
    } else {
	form_clear $w
    }
}

# clear a form
proc form_clear {w} {
    set n [llength [form_get $w]]
    for {set i 0} {$i<$n} {incr i} {
	lappend clear ""
    }
    form_put $w $clear
}

# puts list in the listbox
proc formMenu_put {w list} {
    set w $w.f.sel.list

    $w delete 0 end
    foreach element $list {
	$w insert end $element
    }
}

proc formMenu {w form_w array_var delete_cmd} {
    global $array_var

    frame $w -relief ridge -bd 2
    frame $w.f 
    frame $w.f.sel
    pack [listbox $w.f.sel.list -yscroll [list $w.f.sel.yscroll set] \
	      -selectmode single] -side left
    pack [scrollbar $w.f.sel.yscroll -command [list $w.f.sel.list yview]] \
	-side left -fill y
    pack $w.f.sel
    frame $w.f.b -relief ridge -bd 2
    pack [button $w.f.b.edit -text Edit -command \
	      [list form_active $form_w $w $array_var]] -side left
    pack [button $w.f.b.delete -text Delete -command \
	      [list formMenu_delete $w $delete_cmd $array_var]] -side left
    pack $w.f.b 
    pack $w.f
    
    formMenu_put $w [lsort [array names $array_var]]
    bind $w.f.sel.list <Double-Button-1> [list form_active $form_w $w \
					      $array_var]
    return $w
}

proc form {w labels} {
   frame $w 
   set count 0
   frame $w.f 
   foreach label $labels {
       grid [label $w.f.l$count -text $label] -sticky w -column 0
       grid [entry $w.f.e$count] -row $count -sticky ew -column 1
       incr count
   }
   pack $w.f -fill x -expand 1
   return $w
}
##################################################################

# Dialog boxes


# Construct a dialog box for connecting to a mud either by
# selecting a defined character or typing the location and port in.

# This proc doesn't really fit in with the rest of the array
# editing dialog boxes.
proc dialog_connect {} {
    global MmuclChar Mmucl

    if {[winfo exists .con]} {raise .con;return}
    
    toplevel .con
    wm title .con "Connect"
    
    frame .con.f -relief groove -bd 2
    grid [label .con.f.hostl -text "Host:"]
    grid [entry .con.f.host -textvariable Mmucl(host)] -row 0 -column 1
    grid [label .con.f.portl -text "Port:"] 
    grid [entry .con.f.port -textvariable Mmucl(port)] -row 1 -column 1
    
    if {![info exists Mmucl(char)]} {set Mmucl(char) ""}
    # if characters defined, make a character selection area
    if {[array size MmuclChar]} {
	grid [label .con.f.lchar -text "Character:"]
        eval tk_optionMenu .con.f.char Mmucl(char) \
	    [lsort [array names MmuclChar]]
	balloon_add .con.f.char "Select a character to load"
	grid .con.f.char -row 2 -column 1

	.con.f.char.menu add separator
	.con.f.char.menu add command \
	    -label None -command {set Mmucl(char) ""}
	# How to keep host and port updated from char?
    }

    frame .con.b -relief ridge -bd 2
    pack [button .con.b.ok -text Connect \
	      -command {
		  if {$Mmucl(char)==""} {
		      connect $Mmucl(host) $Mmucl(port)
		  } else {charload $Mmucl(char)}
		  destroy .con}
	 ] -side left
    
    pack [button .con.b.quit -text Cancel -command \
	      {destroy .con}] -side left
  
    pack .con.f .con.b
    wm geometry .con +[winfo pointerx .]+[winfo pointery .]
    wm resizable .con 0 0
}

proc dialog_alias {} {
    
    if {[winfo exists .alias]} {raise .alias;return}
    
    toplevel .alias
    wm title .alias {Edit aliases}
    formMenu .alias.sel .alias.edit.form MmuclAlias unalias
    frame .alias.edit -relief ridge -bd 2
    
    message .alias.edit.msg -text "Aliases execute Tcl commands when they match the first word of input." -aspect 500
    pack .alias.edit.msg -fill x -side top 
        
    pack [frame .alias.edit.f1 -relief ridge -bd 2 -height 2] \
	-fill x -pady 5

    form .alias.edit.form {Name: Commands:}
    pack .alias.edit.form -side top -pady 5
    
    frame .alias.edit.b
    pack [button .alias.edit.b.test -text Test -command \
	      {eval [join [lrange [form_get .alias.edit.form] 1 end]]}] \
	-side left
    balloon_add .alias.edit.b.test {Test the alias commands}
    pack [button .alias.edit.b.add -text Add -command \
	      {form_save .alias.edit.form .alias.sel MmuclAlias alias}] \
	-side left
    balloon_add .alias.edit.b.add {Add to currently defined aliases}
    pack .alias.edit.b -side top
    
    pack [frame .alias.edit.f2 -relief ridge -bd 2 -height 2] \
	-fill x -pady 5
    
    frame .alias.edit.action -bd 2 -relief ridge
    pack [button .alias.edit.action.help -text "Help" \
	      -command {help alias}] \
 	[button .alias.edit.action.exit -text "Exit" \
	     -command {destroy .alias}] -side left
    pack .alias.edit.action -side bottom
    
    pack .alias.sel .alias.edit -side left -fill both
    
    wm geometry .alias +[winfo pointerx .]+[winfo pointery .]
    wm resizable .alias 0 0
}

proc dialog_action {} {

    if {[winfo exists .action]} {raise .action;return}

    toplevel .action
    wm title .action {Edit actions}
    formMenu .action.sel .action.edit.form MmuclAction unaction
    frame .action.edit -relief ridge -bd 2

    message .action.edit.msg -text "Actions execute Tcl commands when mud output matches their pattern." -aspect 500
    pack .action.edit.msg -fill x -side top

    pack [frame .action.edit.f1 -relief ridge -bd 2 -height 2] \
	-fill x -pady 5
    
    form .action.edit.form {Pattern: Commands:}
    pack .action.edit.form -side top -pady 5
    
    frame .action.edit.b
    pack [button .action.edit.b.test -text Test -command \
	      {eval [join [lrange [form_get .action.edit.form] 1 end]]}] \
	-side left
    balloon_add .action.edit.b.test "Test the action commands"
    pack [button .action.edit.b.add -text "Add" \
	      -command {form_save .action.edit.form .action.sel \
			    MmuclAction action}] -side left
    balloon_add .action.edit.b.add "Add to currently defined actions"
    pack .action.edit.b -side top
   
    pack [frame .action.edit.f2 -relief ridge -bd 2 -height 2] \
	-fill x -pady 5
    
    frame .action.edit.action -bd 2 -relief ridge
    pack [button .action.edit.action.help -text "Help" \
	      -command {help action}] \
 	[button .action.edit.action.exit -text "Exit" \
	     -command {destroy .action}] -side left
    pack .action.edit.action -side bottom

    pack .action.sel .action.edit -side left -fill both

    wm geometry .action +[winfo pointerx .]+[winfo pointery .]
    wm resizable .action 0 0
}

proc dialog_character {} {

    if {[winfo exists .character]} {raise .character;return}

    toplevel .character
    wm title .character {Edit characters}
    formMenu .character.sel .character.edit.form MmuclChar chardelete
    frame .character.edit -relief ridge -bd 2

    message .character.edit.msg -text "Register a character." -aspect 1000
    pack .character.edit.msg -fill x -side top 

     pack [frame .character.edit.f1 -relief ridge -bd 2 -height 2] \
	-fill x -pady 5

    form .character.edit.form \
	{Character: "Mud Name:" "Mud Host:" "Mud Port:" Login:}
    pack .character.edit.form -side top -pady 2

    frame .character.edit.b
    pack [button .character.edit.b.add -text "Add" -command \
	      {form_save .character.edit.form .character.sel \
		   MmuclChar charadd}] -side left
    balloon_add .character.edit.b.add {Add to defined characters}
    pack .character.edit.b -side top
  
    pack [frame .character.edit.f2 -relief ridge -bd 2 -height 2] \
	-fill x -pady 5

    frame .character.edit.action -bd 2 -relief ridge
    pack [button .character.edit.action.help -text "Help" \
	      -command {help charadd}] \
 	[button .character.edit.action.exit -text "Exit" \
	     -command {destroy .character}] -side left
    pack .character.edit.action -side bottom

    pack .character.sel .character.edit -side left -fill both

    wm geometry .character +[winfo pointerx .]+[winfo pointery .]
    wm resizable .character 0 0
}

proc dialog_bind {} {
    global bind_array
    
    if {[winfo exists .bind]} {raise .bind;return}

    catch {unset bind_array}
    foreach event [bind .input] {
	set bind_array($event) [list [bind .input $event]]
    }

    toplevel .bind
    wm title .bind {Edit key bindings}
    formMenu .bind.sel .bind.edit.form bind_array bind_delete
    frame .bind.edit -relief ridge -bd 2

    label .bind.edit.msg -text "Bind Tcl commands to key events."
    pack .bind.edit.msg -side top 

    pack [frame .bind.edit.f1 -relief ridge -bd 2 -height 2] -fill x -pady 5

    form .bind.edit.form {Event: Commands:}
    pack .bind.edit.form -side top -pady 5

    frame .bind.edit.b
    pack [button .bind.edit.b.test -text "Test" -command \
	      {eval [join [lrange [form_get .bind.edit.form] 1 end]]}] \
	-side left
    balloon_add .bind.edit.b.test {Test bound commands.}
    pack [button .bind.edit.b.add -text "Add" -command \
	      {form_save .bind.edit.form .bind.sel \
		   bind_array bind_add}] -side left
    balloon_add .bind.edit.b.add {Add to current bindings}
    pack .bind.edit.b -side top

    pack [label .bind.edit.levents -text {Type to see key events}] \
	-side top -pady 5
    entry .bind.edit.event
    bind .bind.edit.event <KeyRelease> {form_put .bind.edit.form <Key-%K>}
    pack .bind.edit.event -side top

    pack [frame .bind.edit.f -relief ridge -bd 2 -height 2] \
	-fill x -pady 5

    frame .bind.edit.action -bd 2 -relief ridge
    pack [button .bind.edit.action.help -text "Help" \
	      -command {help dialog_bind}] \
 	[button .bind.edit.action.exit -text "Exit" \
	     -command {destroy .bind;catch {unset bind_array}}] -side left
    pack .bind.edit.action -side bottom

    pack .bind.sel .bind.edit -side left -fill both

    wm geometry .bind +[winfo pointerx .]+[winfo pointery .]
    wm resizable .bind 0 0
}

##########################################
# Misc stuff.

# Needed for the binding dialog box
proc bind_delete {event} {
    global bind_array

    if {[array names bind_array $event]!=""} {
	bind .input $event {}
	unset bind_array($event)
	echo "** Binding to event $event deleted. **"
    } else {
	echo "** No binding on event $event. **"
    }
}

proc bind_add {event commands} {
    global bind_array
  
    set bind_array($event) [list $commands]
    bind .input $event $commands
    echo "** Event $event bound to $commands **"
}
